"""
This module provides access to pre-trained synthesizability machine
learning models for predicting synthesizability of arbitrary crystal
compounds.
"""


import os
import pandas as pd
import numpy as np
import logging


from monty.serialization import loadfn, dumpfn

from pymatgen.ext.matproj import MPRester

from pymatgen import Composition
import json
from pymatgen import Structure, Lattice
from pumml.learners import PULearner
from matminer.featurizers.structure import DensityFeatures, GlobalSymmetryFeatures
from matminer.featurizers.composition import Meredig, CohesiveEnergy
from pathlib import Path
import tarfile

logging.basicConfig(
    filename="output.log", format="%(levelname)s - %(message)s", level=logging.DEBUG
)


class PUPredict:
    def __init__(self, api_key):
        """Predict synthesizability of all compounds in Materials Project
        Database.

        Features for training data are generated by first-principles
        (density functional theory) calculations, or structural or chemical
        data looked up from Materials Project Database.

        Two models each for F-Block compounds and Non F-Block compounds
        have been pre-trained which are used to predict synthesizability of
        compounds based on the block they belong to.

        Args:
            api_key (str): API key for Materials Project API available from
                https://materialsproject.org/open.

        """

        self.api_key = api_key

        # Extract Data from FigShare
        os.system(
            "wget https://s3-eu-west-1.amazonaws.com/pfigshare-u-files/23873285/pupredict_data.tar.gz"
        )
        self.model_dir = os.getcwd()

        # Extract file from the tar file
        my_tar = tarfile.open("pupredict_data.tar.gz")
        my_tar.extractall("Model_Data")  # specifing which folder to extract to
        my_tar.close()

    def _mpr(self):
        """Get MPRester.

        Returns:
            MPRester (function): A function which is used to extract data from Materials Project

        """
        return MPRester(api_key=self.api_key)

    def _get_mp_id(self, updated_input):
        """Gets the Materials IDs from the database.

        Args:
            updated_input (criteria): The condition on which the data is extracted

        Returns:
            mp_ids (list): List of MP-IDs

        """

        criteria = updated_input

        # get mp-id
        mpids = []
        with self._mpr() as m:
            mp_ids = m.query(criteria, ["material_id"], chunk_size=0)

        mp_ids = [mpid["material_id"] for mpid in mp_ids]

        return mp_ids

    def _get_formula(self, updated_input):
        """Gets the full formula from Materials Project.

        Args:
            updated_input (criteria): The condition on which the data is extracted

        Returns:
            formula (str): A string which contains the formula of the input

        """

        criteria = updated_input

        formula = []

        # get formulae
        with self._mpr() as m:
            formula = m.query(criteria, ["full_formula"], chunk_size=0)

        formula = [id["full_formula"] for id in formula]

        return formula[0]

    def _get_data(self, mp_ids):
        """Gets features of the Materials IDs from the database.

         Args:
             mp_ids (list): List of MP-IDs

         Returns:
             df_input(DataFrame): Pandas DataFrame whcih conatains feature
                from Materials Project Database of the input samples


        """
        chunk_size = 1000
        sublists = [
            mp_ids[i : i + chunk_size] for i in range(0, len(mp_ids), chunk_size)
        ]

        # MPRester.supported_properties
        properties = [
            "theoretical",
            "energy_per_atom",
            "formation_energy_per_atom",
            "e_above_hull",
            "icsd_ids",
            "material_id",
            "structure",
        ]

        data = []
        # Get all materials from MP by mpid
        with self._mpr() as m:
            for sublist in sublists:
                data += m.query(
                    {"material_id": {"$in": sublist}}, properties=properties
                )

        df_input = pd.DataFrame(data)

        # PU_label is 1 (0) if experimental crystal structure exists (doesn't exist)
        df_input["PU_label"] = df_input.theoretical.apply(
            lambda x: 0 if x == True else 1
        )

        return df_input

    def _check_for_block(self, comp):
        """
        Check whether compound belongs to F-Block or Non F-Block.

        Args:
            comp (pymatgen.core.composition.Composition): Contains the
                composition of the the input sample

        Returns:
            fblock (Bool): 0 if it belongs to F-block, 1 if 
                it belongs to non-fblock

        """

        # check which block it belongs to
        if comp.contains_element_type("f-block") == True:
            return 0  # Belongs to F-block
        else:
            return 1  # Belongs to Non F-block

    def _check_for_theoretical(self, updated_input):
        """
        Check if the input is theoretical or experimental and drop the experimental compounds.

        Args:
            updated_input (criteria): The condition on which the data is
                extracted

        Returns:
            mp_ids (list): List of MP-IDs
            df_input(DataFrame): Pandas DataFrame whcih conatains features
                from Materials Project Database of the input samples

        """

        # Get MP-IDs
        mp_id = self._get_mp_id(updated_input)

        # Get Features of input
        df_input = self._get_data(mp_id)

        # Print the MP-ID for experimental compounds
        for idx, elem in enumerate(mp_id):
            if df_input.theoretical.values[idx] == False:
                logging.info(
                    'The given input with material_id "%s" exists.',
                    df_input.material_id.values[idx],
                )

        # Update the df_input by removing experimental compounds
        j = df_input[(df_input.theoretical == False)].index
        df_input = df_input.drop(j)

        return mp_id, df_input

    def _check_if_already_exists(self, comp, updated_input):
        """Check if the input exists in pre-trained model.

        Args:
            updated_input (criteria): The condition on which the data is
                extracted
            comp (pymatgen.core.composition.Composition): Contains the
                composition of the the input sample

        Returns:
            mp_ids (list): List of MP-IDs
            df_input(DataFrame): Pandas DataFrame whcih conatains features
                from Materials Project Database of the input samples
            fblock (Bool): 0 if it belongs to F-block, 1 if it belongs to 
                Non F block
            synth_scores (list): Synthesizability scores (between 0 and 1) of
                given sample.

        """

        mp_id, df_input = self._check_for_theoretical(updated_input)
        fblock = self._check_for_block(comp)

        # Intialise Synthesizability Score
        synth_score = []

        # Load F-Block pre-trained Model
        if fblock == 0:
            fblock_trained = pd.read_json(
                self.model_dir
                + "/Model_Data/0384056001594797986/PUPredict_Data/fblock_trained_model.json"
            )
            mp_id_theo = df_input.material_id.values
            for i in mp_id_theo:
                if i in fblock_trained.material_id.values:
                    synth_score.append(
                        fblock_trained.synth_score[
                            fblock_trained.material_id == i
                        ].values
                    )
                    j = df_input[(df_input.material_id == i)].index
                    df_input = df_input.drop(j)

        # Load Non F-Block pre-trained Model
        elif fblock == 1:
            non_fblock_trained = pd.read_json(
                self.model_dir
                + "/Model_Data/0384056001594797986/PUPredict_Data/non_fblock_trained_model.json"
            )
            mp_id_theo = df_input.material_id.values
            for i in mp_id_theo:
                if i in non_fblock_trained.material_id.values:
                    synth_score.append(
                        non_fblock_trained.synth_score[
                            non_fblock_trained.material_id == i
                        ].values
                    )
                    j = df_input[(df_input.material_id == i)].index
                    df_input = df_input.drop(j)

        return fblock, mp_id, synth_score, df_input

    def _extract_features(df_input):
        """
        Extract features using Matminer from the 'structure' column in
            df_input

         Args:
             df_input (DataFrame): Pandas DataFrame whcih conatains features
                from Materials Project Database of the input samples

         Returns:
             df_extracted (DataFrame): Pandas DataFrame which contains
                features of input samples extracted using Matminer

        """

        # Dropping the 'theoretical' column
        df_input.drop(columns=["theoretical"], inplace=True)

        # Extracting the features
        dfeat = DensityFeatures()
        symmfeat = GlobalSymmetryFeatures()
        mfeat = Meredig()
        cefeat = CohesiveEnergy()

        df_input["density_test"] = df_input.structure.apply(
            lambda x: dfeat.featurize(x)[0]
        )
        df_input["vpa"] = df_input.structure.apply(lambda x: dfeat.featurize(x)[1])
        df_input["packing fraction"] = df_input.structure.apply(
            lambda x: dfeat.featurize(x)[2]
        )
        df_input["spacegroup_num"] = df_input.structure.apply(
            lambda x: symmfeat.featurize(x)[0]
        )
        df_input["cohesive_energy"] = df_input.apply(
            lambda x: cefeat.featurize(
                x.structure.composition,
                formation_energy_per_atom=x.formation_energy_per_atom,
            )[0],
            axis=1,
        )
        df_input["mean AtomicWeight"] = df_input.structure.apply(
            lambda x: mfeat.featurize(x.composition)[-17]
        )
        df_input["range AtomicRadius"] = df_input.structure.apply(
            lambda x: mfeat.featurize(x.composition)[-12]
        )
        df_input["mean AtomicRadius"] = df_input.structure.apply(
            lambda x: mfeat.featurize(x.composition)[-11]
        )
        df_input["range Electronegativity"] = df_input.structure.apply(
            lambda x: mfeat.featurize(x.composition)[-10]
        )
        df_input["mean Electronegativity"] = df_input.structure.apply(
            lambda x: mfeat.featurize(x.composition)[-9]
        )

        # Drop 'structure' column
        df_input.drop(columns=["structure"], inplace=True)

        # ignore compounds that failed to featurize
        df_extracted = df_input.fillna(df_input.mean()).query("cohesive_energy > 0.0")

        # Re-arranging the 'PU Label' column
        pu_label = df_extracted["PU_label"]
        df_extracted = df_extracted.drop(["PU_label"], axis=1)
        df_extracted["PU_label"] = pu_label

        # Drop the icsd_ids column
        df_extracted.drop(columns=["icsd_ids"], inplace=True)

        return df_extracted

    def _append_data(self, df_extracted, fblock):
        """Append featurized data.

        Args:
            df_extracted (DataFrame): Pandas DataFrame which contains features
                of input samples extracted using Matminer
            fblock (Bool): 0 if it belongs to F-block, 1 if it
                belongs to Non F block

        Returns:
            df (DataFrame): Pandas DataFrame which contains features of
                input samples extracted using Matmimer along with training
                data
            input_ids(numpy array): MP-IDs of updated input compounds

        """

        # append data to the respective block
        input_ids = df_extracted.material_id.values

        # If it belongs to 'F-block'
        if fblock == 0:
            # F-Block training Dataset
            fblock_features = pd.read_json(
                self.model_dir
                + "/Model_Data/0384056001594797986/PUPredict_Data/fblock_extracted_features.json"
            )
            df = fblock_features.append(df_extracted)
            df.reset_index(drop=True, inplace=True)

        # If it belongs to 'Non F-Block'
        else:
            # Non F-Block training dataset
            non_fblock_features = pd.read_json(
                self.model_dir
                + "/Model_Data/0384056001594797986/PUPredict_Data/non_fblock_extracted_features.json"
            )
            df = non_fblock_features.append(df_extracted)
            df.reset_index(drop=True, inplace=True)

        return input_ids, df

    def _do_pulearner(input_ids, df):
        """Train the model using PULearner.

        Args:
        df (DataFrame): Pandas DataFrame which contains features of input
            samples extracted using Matmimer along with training data
        input_ids (numpy array): MP-IDs of updated input compounds


        Returns:
            synth_scores (list): Synthesizability scores (between 0 and 1) of 
                given sample.

        """

        # Run PUMML
        df.to_json("test.json")
        pul = PULearner()

        # Set hyperparameters
        n_splits = 10  # kfold CV
        n_repeats = 3  # Repeat the entire kfold CV n times for averaging
        n_bags = 100  # bags for bootstrap aggregating.

        pu_stats = pul.cv_baggingDT(
            "test.json", splits=n_splits, repeats=n_repeats, bags=n_bags
        )
        df1 = pul.df_U.copy()
        df1["synth_score"] = pu_stats["prob"]

        # Initialsing synth_score
        synth_score = []
        for i in range(len(input_ids)):
            synth_score.append(df1.synth_score[df1.material_id == input_ids[i]].values)

        return synth_score

    def synth_score_from_mpid(self, mpid):
        """
        Takes input in form of Materials Project ID and returns
        the synthesizability score for the same.

        Args:
            mpid (str): For example, 'mp-1234'

         Returns:
            synth_scores (list): Synthesizability scores (between 0 and 1) of
                given sample.

        """

        # Updating the criteria
        updated_input = {"material_id": mpid}

        # Extracting formula
        formula = self._get_formula(updated_input)

        # Extracting composition
        comp = Composition(formula)

        # Extracting MP-IDs
        mp_id = self._get_mp_id(updated_input)

        # If Input given is invalid
        if len(mp_id) == 0:
            return "No such compound exists in Materials Project Database"

        # Checking if it already exists in pre-trained model
        fblock, mp_id, synth_score, df_input = self._check_if_already_exists(
            comp, updated_input
        )

        # If found in pre-trained models, report synth_scores
        if df_input.shape[0] == 0:
            if len(synth_score) != 0:
                logging.info("The synthesizability scores are %s", synth_score)
                return synth_score

        # Else, train the model and report scores
        else:
            df_extracted = self._extract_features(df_input)
            input_ids, df = self._append_data(df_extracted, fblock)
            score = self._do_pulearner(input_ids, df)
            synth_score.append(score)
            logging.info("The synthesizability scores are %s", synth_score)
            return synth_score

    def synth_score_from_formula(self, formula):
        """
        Takes input in form of an unreduced crystal composition and returns
        the synthesizability score for the same.

        Args:
            formula (str): For example: 'Ba2Yb2Al4Si2N10O4'

         Returns:
            synth_scores (list): Synthesizability scores (between 0 and 1) of
                 given sample.

        """

        # Updating the criteria
        updated_input = {"full_formula": formula}

        # Extracting the composition
        comp = Composition(formula)

        # Extracting the MP-IDs
        mp_id = self._get_mp_id(updated_input)

        # If Input given is invalid
        if len(mp_id) == 0:
            return "No such compound exists in Materials Project Database"

        # Checking if it already exists in pre-trained model
        fblock, mp_id, synth_score, df_input = self._check_if_already_exists(
            comp, updated_input
        )

        # If found in pre-trained models, report synth_scores
        if df_input.shape[0] == 0:
            if len(synth_score) != 0:
                logging.info("The synthesizability scores are %s", synth_score)
                return synth_score

        # Else, train the model and report scores
        else:
            df_extracted = self._extract_features(df_input)
            input_ids, df = self._append_data(df_extracted, fblock)
            score = self._do_pulearner(input_ids, df)
            synth_score.append(score)
            logging.info("The synthesizability scores are %s", synth_score)
            return synth_score

    def synth_score_from_structure(self, structure):
        """
        Takes input in form of Pymatgen Structure Object and returns the
        synthesizability score for the same.

        Args:
            structure (pymatgen.core.structure.Structure): structure input.

        Returns:
            synth_scores (list): Synthesizability scores (between 0 and 1) of
                given sample.

        """

        # Extract formula and remove the spaces between them
        formula = structure.formula
        formula = formula.replace(" ", "")

        synth_score = self.synth_score_from_formula(formula)
        return synth_score


"""
Deploy semi-supervised PU machine learning models.

This module provides classes for training, testing, and deploying a PU
learning model for predicting material synthesizability. Utility functions
for plotting aid in visualizing and analyzing results.

References:
    [1] DOI: 10.1021/acsnano.8b08014
    [2] DOI: 10.1145/1401890.1401920
    [3] DOI: 10.1016/j.patrec.2013.06.010
"""

from sklearn.ensemble import RandomForestClassifier
from sklearn.metrics import precision_recall_fscore_support
from sklearn.cluster import KMeans
from sklearn.mixture import GaussianMixture, BayesianGaussianMixture
from sklearn.tree import DecisionTreeClassifier
from sklearn.metrics import precision_recall_curve
from sklearn.model_selection import RepeatedKFold
from sklearn.utils import resample

from mpl_toolkits.mplot3d import Axes3D

from monty.serialization import dumpfn

import pandas as pd
import seaborn as sns
import os
import pickle

import numpy as np
import matplotlib.pyplot as plt
from pylab import rcParams

__author__ = "Nathan C. Frey, Jin Wang"
__copyright__ = "MIT License"
__version__ = "0.0.1"
__maintainer__ = "Nathan C. Frey"
__email__ = "n.frey@seas.upenn.edu"
__status__ = "Development"
__date__ = "Aug 2017"


class PULearner:
    def __init__(self):
        """A machine learning model that predicts material synthesizability.

        Positive samples are experimentally synthesized materials. Unlabeled
        samples are not-yet synthesized materials.

        Features for training data might be generated by first-principles 
        (density functional theory) calculations, or structural or chemical
        data looked up from a table.

        Hyperparameters are initialized with sensible defaults, but any newly
        trained model should have hyperparams carefully converged.

        Attributes:
            pu_stats (dict): Outputs of cv_baggingDT
            df_U (DataFrame): Unlabeled data.
            df_P (DataFrame): Positive data.

            synth_scores (list): Synthesizability scores (between 0 and 1) of
                unlabeled samples.
            labels (list): Likely synthesizable (1) or not (0)
            feat_importances (DataFrame): Feature importances from trained
                decision tree classifiers. Index corresponds to feature index
                in original data. 

        """

        self.pu_stats = None
        self.df_U = None
        self.df_P = None
        self.synth_scores = None
        self.labels = None
        self.feat_importances = None

    def cv_baggingDT(self, pu_data, splits=10, repeats=10, bags=100, filename=""):
        """
        Train bagged decision tree base classifiers and do repeated 
        k-fold CV.

        Synthesizability scores (0 = not synthesizable, 1 = already
        synthesized) are generated for an unlabeled sample by averaging
        the scores from the ensemble of decision tree classifiers that
        have not been trained on that sample. 

        Args:
            pu_data (json): A file where each row describes a material.
                There MUST be a column called "PU_label" where a 1 value
                indicates a synthesized (positive) compound and a 0 value
                indicates an unlabeled compound.

            splits (int): Number of splits in k-fold CV.
            repeats (int): Number of repeated k-fold CV.
            bags (int): Number of bags in bootstrap aggregation.
            filename (string): Save model training results to file with
                filename ending in .json or .pkl.

        Returns:
            pu_stats (dict): Metrics and outputs of PU learning model
                training.

        """
        
        print("Start PU Learning.")

        # Preprocess data and set attributes
        df = pd.read_json(pu_data)
        df_P, df_U, X_P, X_U = self._process_pu_data(df)
        self.df_P = df_P
        self.df_U = df_U

        # Split data into training and test splits for k-fold CV
        kfold = RepeatedKFold(n_splits=splits, n_repeats=repeats, random_state=42)

        # Scores for PU learning (tpr = True Positive Rate)
        scores = []
        tprs = []

        # Predicted synthesis probability of CVed P and U sets
        prob_P = np.ones(shape=(X_P.shape[0], splits * repeats))
        prob_U = -np.ones(shape=(X_U.shape[0], splits * repeats))

        # Feature importance
        feat_rank = np.zeros(shape=(X_P.shape[1], splits * repeats))

        idsp = 0  # index of repeated k splits

        # Loop over P and U training/test samples
        for (ptrain, ptest), (utrain, utest) in zip(kfold.split(X_P), kfold.split(X_U)):

            # Number of P and U training samples
            N_ptrain = X_P[ptrain].shape[0]
            N_utrain = X_U[utrain].shape[0]

            d = X_P.shape[1]
            K = N_ptrain
            train_label = np.zeros(shape=(N_ptrain + K,))
            train_label[:N_ptrain] = 1.0  # Synthesized (positive)

            # Out of bag samples
            n_oob = np.zeros(shape=(N_utrain,))
            f_oob = np.zeros(shape=(N_utrain, 2))

            # Sums of probabilities of test sets
            f_ptest = np.zeros(shape=(X_P[ptest].shape[0], 2))
            f_utest = np.zeros(shape=(X_U[utest].shape[0], 2))

            # Bootstrap resampling for each bag
            for i in range(bags):
                bootstrap_sample = np.random.choice(
                    np.arange(N_utrain), replace=True, size=K
                )

                # Positive samples and bootstrapped unlabeled samples
                data_bootstrap = np.concatenate(
                    (X_P[ptrain], X_U[bootstrap_sample, :]), axis=0
                )

                # Train decision tree classifier
                model = DecisionTreeClassifier(
                    max_depth=None,
                    max_features=None,
                    criterion="gini",
                    class_weight="balanced",
                )

                model.fit(data_bootstrap, train_label)

                # Index for the oob samples
                idx_oob = sorted(
                    set(range(N_utrain)) - set(np.unique(bootstrap_sample))
                )

                # Transductive learning on oob samples
                f_oob[idx_oob] += model.predict_proba(X_U[utrain][idx_oob])
                n_oob[idx_oob] += 1
                f_ptest += model.predict_proba(X_P[ptest])
                f_utest += model.predict_proba(X_U[utest])
                feat_rank[:, idsp] = model.feature_importances_

            # Predicted synthesis probabilities of unlabeled samples
            predict_utrain = f_oob[:, 1] / n_oob

            # Predicted probabilities for P and U test sets
            predict_ptest = f_ptest[:, 1] / bags
            predict_utest = f_utest[:, 1] / bags

            # Find predicted positives
            true_pos = predict_ptest[np.where(predict_ptest > 0.5)].shape[0]
            u_pos = predict_utest[np.where(predict_utest > 0.5)].shape[0]

            N_ptest = X_P[ptest].shape[0]
            N_utest = X_U[utest].shape[0]

            # Predicted positive ratio in test set
            p_pred_pos = (true_pos + u_pos) / (N_ptest + N_utest) + 0.0001

            # Compute PU recall (TPR) and score metrics
            recall = true_pos / N_ptest
            score = recall ** 2 / p_pred_pos
            scores.append(score)
            tprs.append(recall)

            # Predicted probabilities
            prob_P[ptest, idsp] = predict_ptest
            prob_U[utrain, idsp] = predict_utrain
            prob_U[utest, idsp] = predict_utest
            idsp += 1

            # Progress update
            if (idsp + 1) % splits == 0:
                tpr_tmp = np.asarray(tprs[-splits - 1 : -1])
                print(
                    "Performed Repeated "
                    + str(splits)
                    + "-fold: "
                    + str(idsp // splits + 1)
                    + " out of "
                    + str(repeats)
                )
                print(
                    "True Positive Rate: %0.2f (+/- %0.2f)"
                    % (tpr_tmp.mean(), tpr_tmp.std() * 2)
                )

        # Predicted labels from k-fold CV
        label_U = np.zeros(shape=(X_U.shape[0], splits * repeats + 1), dtype=int)
        label_U[:, : splits * repeats][np.where(prob_U > 0.5)] = 1
        label_U[:, splits * repeats] = np.sum(
            label_U[:, : splits * repeats + 1], axis=1
        )

        tprs = np.asarray(tprs)
        scores = np.asarray(scores)

        # Metrics for each model in the k-folds
        label_U_rp = np.zeros(shape=(X_U.shape[0], repeats), dtype=int)
        prob_U_rp = np.zeros(shape=(X_U.shape[0], repeats))
        feat_rank_rp = np.zeros(shape=(X_U.shape[1], repeats))
        tpr_rp = np.zeros(shape=(repeats,))
        scores_rp = np.zeros(shape=(repeats,))
        labels = np.zeros(shape=(X_U.shape[0],))

        for i in range(repeats):
            prob_U_rp[:, i] = prob_U[:, i * splits : (i + 1) * splits].mean(axis=1)
            feat_rank_rp[:, i] = feat_rank[:, i * splits : (i + 1) * splits].mean(
                axis=1
            )
            tpr_rp[i] = tprs[i * splits : (i + 1) * splits].mean()
            scores_rp[i] = scores[i * splits : (i + 1) * splits].mean()

        label_U_rp[np.where(prob_U_rp > 0.5)] = 1
        prob = prob_U_rp.mean(axis=1)
        labels[np.where(prob > 0.5)] = 1

        # Get confidence interval of TPR for each kfold
        tpr_low, tpr_up = self.bootstrapCI(tpr_rp)
        scores_low, scores_up = self.bootstrapCI(scores_rp)

        # PU learning metrics
        metrics = np.asarray(
            [tpr_rp.mean(), tpr_low, tpr_up, scores_rp.mean(), scores_low, scores_up]
        )

        print("Accuracy: %0.2f" % (tpr_rp.mean()))
        print("95%% confidence interval: [%0.2f, %0.2f]" % (tpr_low, tpr_up))

        # Metrics and results from training / testing
        pu_stats = {
            "prob": prob,
            "labels": labels,
            "metrics": metrics,
            "prob_rp": prob_U_rp,
            "label_rp": label_U_rp,
            "tpr_rp": tpr_rp,
            "scores_rp": scores_rp,
            "feat_rank_rp": feat_rank_rp,
        }

        # Save results
        if filename:
            if filename.endswith(".json"):
                dumpfn(pu_stats, filename)
            if filename.endswith(".pkl"):
                with open(filename, "wb") as file:
                    pickle.dump(pu_stats, file, protocol=pickle.HIGHEST_PROTOCOL)

        self.pu_stats = pu_stats
        return pu_stats

    def bootstrapCI(self, data, ci=95, ns=10000):
        """Compute confidence interval of the TPR.

        Args:
            data (array): Array of TPRs for each kfold.
            ci (int): Confidence interval.
            ns (int): Number of bootstrap resamplings.

        Returns:
            lower (float): Lower endpoint of CI.
            upper (float): Upper endpoint of CI.
            
        """

        bs_rsample = []
        for _ in range(ns):
            rsample = resample(data, n_samples=len(data))
            bs_rsample.append(np.mean(rsample))

        bs_rsample = np.asarray(bs_rsample)
        lower = np.percentile(bs_rsample, (100 - ci) / 2)
        upper = np.percentile(bs_rsample, ci + (100 - ci) / 2)

        return lower, upper

    def corr_heatmap(self, num_feats=10, fname=""):
        """Plot correlation matrix between synthesizability and features.

        cv_baggingDT must be run first.

        Args:
            num_feats (int): How many features to consider.
            fname (str): Filename if correlation plot should be saved.

        Returns:
            None (generates plots)

        """

        pu_stats = self.pu_stats
        df_U = self.df_U
        df_U_copy = df_U.drop(columns=["PU_label"])

        # Get normalized, sorted & ranked list of most important features
        synth_scores = pu_stats["prob"]
        df_U_copy["synth_score"] = synth_scores

        # Make correlation matrix of top "num_feats" features
        corrmat = df_U_copy.corr()
        cols = corrmat.nlargest(num_feats, "synth_score")["synth_score"].index
        cm = np.corrcoef(df_U_copy[cols].values.T)

        sns.set(style='ticks')
        rcParams['figure.dpi'] = 300

        fig, ax = plt.subplots(1, 1)
        hm = sns.heatmap(
            cm,
            ax=ax,
            cbar=True,
            annot=True,
            square=True,
            fmt=".2f",
            annot_kws={"size": 7},
            yticklabels=cols.values,
            xticklabels=cols.values,
        )

        if fname:
            self.save_plot(fname + ".png", fig, ax)

    def get_feat_importances(self, plot_format=""):
        """Process output from PU learning k-fold cross validation.

        cv_baggingDT must be run first.

        If plot_format is specified, a feature importance plot will
        be saved.

        Args:
            plot_format (str): svg, png, or pdf file format for saving simple
                visualizations of feature importance and correlation. 

        """

        pu_stats = self.pu_stats

        # Feature importances for individual repetitions of kfold CV
        feat_rank_rp = pu_stats["feat_rank_rp"]
        feat_importances = np.sum(feat_rank_rp, axis=1)

        df_U = self.df_U
        df_U = df_U._get_numeric_data()
        df_U_copy = df_U.drop(columns=["PU_label"])
        feat_names = df_U_copy.columns

        # Index corresponds to feature in original data
        df_feat = pd.DataFrame(columns=["feature", "importance"])
        df_feat["feature"] = feat_names
        df_feat["importance"] = feat_importances

        # Sort by importance
        df_feat_sort = df_feat.sort_values(by="importance", ascending=False)
        max_value = df_feat["importance"].max()

        # Normalize to 1
        df_feat_sort["importance"] = df_feat_sort["importance"] / max_value

        # Set feature importance attribute
        self.feat_importances = df_feat

        if plot_format in ["svg", "pdf", "png"]:

            # Feature importance plot
            fig, ax = plt.subplots(figsize=(10, 4))
            with sns.axes_style(style="ticks"):
                sns.barplot(x="feature", y="importance", data=df_feat_sort)
            ax.set_xticklabels(
                ax.get_xticklabels(), rotation=45, ha="right", fontsize=7
            )
            filename = "feat_importance." + plot_format
            self.save_plot(filename, fig, ax)

    @staticmethod
    def _process_pu_data(data):
        """Utility method for processing input data.

        Args:
            data (DataFrame): Data with positive and unlabeled samples.

        Returns:
            X_P (array): Positive sample set.
            X_U (array): Unlabeled sample set.

        """

        df_P = data.query("PU_label == 1")  # Positive value is 1
        df_U = data.query("PU_label == 0")  # Unlabeled value is 0

        # Chop off PU label and drop non-numeric columns for sklearn
        X_P = np.asarray(df_P.drop(columns=["PU_label"])._get_numeric_data())
        X_U = np.asarray(df_U.drop(columns=["PU_label"])._get_numeric_data())

        return df_P, df_U, X_P, X_U

    @staticmethod
    def save_plot(filename, fig, ax):
        """Utility method for saving simple visualizations.

        Args:
            filename (str): Name ending in .svg, .png, or .pdf
            fig, ax (objects): Matplotlib objects.

        Returns:
            None

        """

        sns.set_style("ticks")
        fig.tight_layout()
        fig.savefig(filename)


class PUInteract:
    def __init__(self, df_parent, pu_parent, df_child, pu_child, merge_on=(), feats=()):
        """Consider parent and child phase PU learning scores.

        This class looks at PU learning scores for parent bulk
        compounds (e.g. layered h-BN) and scores of the child phases
        along with descriptors like exfoliation energy and changes
        in structural/electronic properties to predict (parent, child)
        pairs that can be synthesized.

        Parent and child must be linked by a column that allows the
        dataframes to be merged. There should also be additional features
        that characterize the structural and chemical differences between
        parents and children, e.g. changes in bond lengths, etc.

        Unsupervised clustering models are used to identify synthesizable 
        (parent/child) pairs.

        Args:
            df_parent (str): Parent data filename.
            pu_parent (dict): Output from PULearner.cv_baggingDT.
            df_child (str): Child data filename.
            pu_child (dict): Output from PULearner.cv_baggingDT.
            merge_on (tuple): Column name(s) on which to merge.
            feats (tuple): Column names to use as features. If empty, use all
                possible columns. 

        Attributes:
            merged_df (DataFrame): (Parent, child) pair data.
            X (array): Array representation of merged_df.

        Returns:
            None

        """

        df_parent = pd.read_json(df_parent)
        df_child = pd.read_json(df_child)

        # Set scores from PULearner
        df_parent["synth_score"] = 1
        df_child["synth_score"] = 1

        df_parent.loc[df_parent.eval("PU_label == 0"), "synth_score"] = pu_parent[
            "prob"
        ]
        df_child.loc[df_child.eval("PU_label == 0"), "synth_score"] = pu_child["prob"]

        # Merge parent and child dfs
        merge_on = list(merge_on)
        df = pd.merge(
            df_parent, df_child, on=merge_on, how="outer", suffixes=["_p", "_c"]
        )
        df.drop(columns=["PU_label_p", "PU_label_c"], inplace=True, axis=1)

        if feats:
            feat_names = [f + "_p" for f in feats] + [f + "_c" for f in feats]
            df = df[feat_names]

        self.merged_df = df
        self.X = np.array(df)

    def do_kmeans(self, n_clusters=2, seed=42):
        """Do k-means clustering on (parent, child) pairs.

        Args:
            n_clusters (int): Number of clusters.
            seed (int): Fix random seed for kmeans reproducibility.

        Returns:
            kmeans_output (dict): kmeans cluster centers, cluster labels for
                each (parent, child)

        """

        np.random.seed(seed)
        km = KMeans(n_clusters=n_clusters, random_state=seed)

        km.fit(self.X)
        kmeans_output = {
            "cluster_centers": km.cluster_centers_,
            "cluster_labels": km.labels_,
        }

        return kmeans_output

    def do_gmixture(self, n_components=2, seed=42):
        """
        Estimate parameters of a Gaussian mixture distribution of (parent,
            child) data.

        Args:
            n_components (int): Number of components in GMM.
            seed (int): Random seed.

        Returns:
            gmm_output (dict): Predicted labels of (parent, child) pairs and
                predicted posterior probabilities of each component.

        """

        np.random.seed(seed)
        gmm = GaussianMixture(
            n_components=n_components, random_state=seed, covariance_type="full"
        )

        gmm.fit(self.X)
        gmm_labels = gmm.predict(self.X)
        gmm_prob = gmm.predict_proba(self.X)[:, 0]
        gmm_output = {"gmm_labels": gmm_labels, "gmm_prob": gmm_prob}

        return gmm_output

    def do_bgm(self, n_components=6, seed=42):
        """Bayesian Gaussian Mixture.

        Infer the effective number of components in a Gaussian Mixture Model
            via variational Bayesian estimation.

        n_effective_componenents < n_components if the model sets some
            weights close to 0.

        Args:
            n_components (int): Number of components in GMM.
            seed (int): Random seed.

        Returns:
            bgm_output (dict): Labels and probabilities.

        """

        np.random.seed(seed)
        bgm = BayesianGaussianMixture(
            n_components=n_components,
            covariance_type="full",
            weight_concentration_prior=1e-2,
            weight_concentration_prior_type="dirichlet_process",
            mean_precision_prior=1e-2,
            init_params="random",
            max_iter=100,
            random_state=seed,
        )

        bgm.fit(self.X)
        bgm_labels = bgm.predict(self.X)
        bgm_prob = bgm.predict_proba(self.X)[:, 0]

        bgm_output = {"bgm_labels": bgm_labels, "bgm_prob": bgm_prob}

        return bgm_output



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pumml",
    version="0.0.2",
    author="Nathan C. Frey, Jin Wang",
    maintainer="Nathan C. Frey",
    author_email="n.frey@seas.upenn.edu",
    description="Positive and Unlabeled Materials Machine Learning (pumml) is a code that uses semi-supervised positive and unlabeled (PU) machine learning to classify materials when data is incomplete and only examples of 'positive' materials are available.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ncfrey/pumml",
    packages=setuptools.find_packages(),
    python_requires='>=3.6',
    install_requires=['matplotlib>=3.1.1', 'monty>=2.0.4', 'numpy==1.19.4', 'pandas==1.1.4', 'scikit-learn==0.23.2', 'scipy>=1.3.0', 'seaborn>=0.9.0',
        'matminer==0.6.3',
        'pymatgen==2020.11.11'],
    keywords=["VASP", "machine", "learning", "materials", "science", "DFT", "synthesis"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Scientific/Engineering :: Physics",
        "Topic :: Scientific/Engineering :: Chemistry",
        "Topic :: Software Development :: Libraries :: Python Modules"
    ],
)
